@@ -2,7 +2,6 @@ package generators
 
 import (
 	"context"
-	"github.com/sirupsen/logrus"
 	"testing"
 	"time"
 
@@ -14,11 +13,12 @@ import (
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	"sigs.k8s.io/controller-runtime/pkg/client/fake"
 
+	"github.com/argoproj/argo-cd/v2/applicationset/services/mocks"
+
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/mock"
 	apiextensionsv1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"
 
-	testutils "github.com/argoproj/argo-cd/v2/applicationset/utils/test"
 	argoprojiov1alpha1 "github.com/argoproj/argo-cd/v2/pkg/apis/application/v1alpha1"
 )
 
@@ -31,7 +31,7 @@ func TestMatrixGenerate(t *testing.T) {
 	}
 
 	listGenerator := &argoprojiov1alpha1.ListGenerator{
-		Elements: []apiextensionsv1.JSON{{Raw: []byte(`{"cluster": "Cluster","url": "Url"}`)}},
+		Elements: []apiextensionsv1.JSON{{Raw: []byte(`{"cluster": "Cluster","url": "Url", "templated": "test-{{path.basenameNormalized}}"}`)}},
 	}
 
 	testCases := []struct {
@@ -51,8 +51,8 @@ func TestMatrixGenerate(t *testing.T) {
 				},
 			},
 			expected: []map[string]interface{}{
-				{"path": "app1", "path.basename": "app1", "path.basenameNormalized": "app1", "cluster": "Cluster", "url": "Url"},
-				{"path": "app2", "path.basename": "app2", "path.basenameNormalized": "app2", "cluster": "Cluster", "url": "Url"},
+				{"path": "app1", "path.basename": "app1", "path.basenameNormalized": "app1", "cluster": "Cluster", "url": "Url", "templated": "test-app1"},
+				{"path": "app2", "path.basename": "app2", "path.basenameNormalized": "app2", "cluster": "Cluster", "url": "Url", "templated": "test-app2"},
 			},
 		},
 		{
@@ -271,6 +271,28 @@ func TestMatrixGenerateGoTemplate(t *testing.T) {
 				{"a": "2", "b": "2"},
 			},
 		},
+		{
+			name: "parameter override: first list elements take precedence",
+			baseGenerators: []argoprojiov1alpha1.ApplicationSetNestedGenerator{
+				{
+					List: &argoprojiov1alpha1.ListGenerator{
+						Elements: []apiextensionsv1.JSON{
+							{Raw: []byte(`{"booleanFalse": false, "booleanTrue": true, "stringFalse": "false", "stringTrue": "true"}`)},
+						},
+					},
+				},
+				{
+					List: &argoprojiov1alpha1.ListGenerator{
+						Elements: []apiextensionsv1.JSON{
+							{Raw: []byte(`{"booleanFalse": true, "booleanTrue": false, "stringFalse": "true", "stringTrue": "false"}`)},
+						},
+					},
+				},
+			},
+			expected: []map[string]interface{}{
+				{"booleanFalse": false, "booleanTrue": true, "stringFalse": "false", "stringTrue": "true"},
+			},
+		},
 		{
 			name: "returns error if there is less than two base generators",
 			baseGenerators: []argoprojiov1alpha1.ApplicationSetNestedGenerator{
@@ -404,6 +426,10 @@ func TestMatrixGetRequeueAfter(t *testing.T) {
 
 	pullRequestGenerator := &argoprojiov1alpha1.PullRequestGenerator{}
 
+	scmGenerator := &argoprojiov1alpha1.SCMProviderGenerator{}
+
+	duckTypeGenerator := &argoprojiov1alpha1.DuckTypeGenerator{}
+
 	testCases := []struct {
 		name               string
 		baseGenerators     []argoprojiov1alpha1.ApplicationSetNestedGenerator
@@ -461,6 +487,30 @@ func TestMatrixGetRequeueAfter(t *testing.T) {
 			},
 			expected: time.Duration(30 * time.Minute),
 		},
+		{
+			name: "returns the default time for duck type generator",
+			baseGenerators: []argoprojiov1alpha1.ApplicationSetNestedGenerator{
+				{
+					Git: gitGenerator,
+				},
+				{
+					ClusterDecisionResource: duckTypeGenerator,
+				},
+			},
+			expected: time.Duration(3 * time.Minute),
+		},
+		{
+			name: "returns the default time for scm generator",
+			baseGenerators: []argoprojiov1alpha1.ApplicationSetNestedGenerator{
+				{
+					Git: gitGenerator,
+				},
+				{
+					SCMProvider: scmGenerator,
+				},
+			},
+			expected: time.Duration(30 * time.Minute),
+		},
 	}
 
 	for _, testCase := range testCases {
@@ -471,18 +521,22 @@ func TestMatrixGetRequeueAfter(t *testing.T) {
 
 			for _, g := range testCaseCopy.baseGenerators {
 				gitGeneratorSpec := argoprojiov1alpha1.ApplicationSetGenerator{
-					Git:         g.Git,
-					List:        g.List,
-					PullRequest: g.PullRequest,
+					Git:                     g.Git,
+					List:                    g.List,
+					PullRequest:             g.PullRequest,
+					SCMProvider:             g.SCMProvider,
+					ClusterDecisionResource: g.ClusterDecisionResource,
 				}
 				mock.On("GetRequeueAfter", &gitGeneratorSpec).Return(testCaseCopy.gitGetRequeueAfter, nil)
 			}
 
 			var matrixGenerator = NewMatrixGenerator(
 				map[string]Generator{
-					"Git":         mock,
-					"List":        &ListGenerator{},
-					"PullRequest": &PullRequestGenerator{},
+					"Git":                     mock,
+					"List":                    &ListGenerator{},
+					"PullRequest":             &PullRequestGenerator{},
+					"SCMProvider":             &SCMProviderGenerator{},
+					"ClusterDecisionResource": &DuckTypeGenerator{},
 				},
 			)
 
@@ -849,7 +903,7 @@ func TestMatrixGenerateListElementsYaml(t *testing.T) {
 	}
 
 	listGenerator := &argoprojiov1alpha1.ListGenerator{
-		Elements: []apiextensionsv1.JSON{},
+		Elements:     []apiextensionsv1.JSON{},
 		ElementsYaml: "{{ .foo.bar | toJson }}",
 	}
 
@@ -871,60 +925,59 @@ func TestMatrixGenerateListElementsYaml(t *testing.T) {
 			},
 			expected: []map[string]interface{}{
 				{
-					"chart":         "a",
-					"version":         "1",
+					"chart":   "a",
+					"version": "1",
 					"foo": map[string]interface{}{
 						"bar": []interface{}{
 							map[string]interface{}{
-								"chart": "a",
+								"chart":   "a",
 								"version": "1",
 							},
 							map[string]interface{}{
-								"chart": "b",
+								"chart":   "b",
 								"version": "2",
 							},
 						},
 					},
 					"path": map[string]interface{}{
-						"basename": "dir",
+						"basename":           "dir",
 						"basenameNormalized": "dir",
-						"filename": "file_name.yaml",
+						"filename":           "file_name.yaml",
 						"filenameNormalized": "file-name.yaml",
-						"path": "path/dir",
-						"segments": []string {
+						"path":               "path/dir",
+						"segments": []string{
 							"path",
 							"dir",
 						},
 					},
 				},
 				{
-					"chart":         "b",
-					"version":         "2",
+					"chart":   "b",
+					"version": "2",
 					"foo": map[string]interface{}{
 						"bar": []interface{}{
 							map[string]interface{}{
-								"chart": "a",
+								"chart":   "a",
 								"version": "1",
 							},
 							map[string]interface{}{
-								"chart": "b",
+								"chart":   "b",
 								"version": "2",
 							},
 						},
 					},
 					"path": map[string]interface{}{
-						"basename": "dir",
+						"basename":           "dir",
 						"basenameNormalized": "dir",
-						"filename": "file_name.yaml",
+						"filename":           "file_name.yaml",
 						"filenameNormalized": "file-name.yaml",
-						"path": "path/dir",
-						"segments": []string {
+						"path":               "path/dir",
+						"segments": []string{
 							"path",
 							"dir",
 						},
 					},
 				},
-
 			},
 		},
 	}
@@ -953,27 +1006,26 @@ func TestMatrixGenerateListElementsYaml(t *testing.T) {
 					"foo": map[string]interface{}{
 						"bar": []interface{}{
 							map[string]interface{}{
-								"chart": "a",
+								"chart":   "a",
 								"version": "1",
 							},
 							map[string]interface{}{
-								"chart": "b",
+								"chart":   "b",
 								"version": "2",
 							},
 						},
 					},
 					"path": map[string]interface{}{
-						"basename": "dir",
+						"basename":           "dir",
 						"basenameNormalized": "dir",
-						"filename": "file_name.yaml",
+						"filename":           "file_name.yaml",
 						"filenameNormalized": "file-name.yaml",
-						"path": "path/dir",
-						"segments": []string {
+						"path":               "path/dir",
+						"segments": []string{
 							"path",
 							"dir",
 						},
 					},
-
 				}}, nil)
 				genMock.On("GetTemplate", &gitGeneratorSpec).
 					Return(&argoprojiov1alpha1.ApplicationSetTemplate{})
@@ -1006,273 +1058,6 @@ func TestMatrixGenerateListElementsYaml(t *testing.T) {
 	}
 }
 
-func TestSkipInterpolatedMatrixCalls(t *testing.T) {
-	interpolatedGitGenerator := &argoprojiov1alpha1.GitGenerator{
-		RepoURL:  "RepoURL",
-		Revision: "Revision",
-		Files: []argoprojiov1alpha1.GitFileGeneratorItem{
-			{Path: "examples/git-generator-files-discovery/cluster-config/{{name}}.json"},
-		},
-	}
-
-	nonInterpolatedGitGenerator := &argoprojiov1alpha1.GitGenerator{
-		RepoURL:  "RepoURL",
-		Revision: "Revision",
-		Files: []argoprojiov1alpha1.GitFileGeneratorItem{
-			{Path: "examples/git-generator-files-discovery/cluster-config/base.json"},
-		},
-	}
-
-	interpolatedClusterGenerator := &argoprojiov1alpha1.ClusterGenerator{
-		Selector: metav1.LabelSelector{
-			MatchLabels: nil,
-			MatchExpressions: []metav1.LabelSelectorRequirement{
-				{
-					Key:      "environment",
-					Operator: "Exists",
-					Values:   []string{},
-				},
-			},
-		},
-	}
-	testCases := []struct {
-		name             string
-		baseGenerators   []argoprojiov1alpha1.ApplicationSetNestedGenerator
-		expectedErr      error
-		expected         []map[string]interface{}
-		clientError      bool
-		expectedMock     [][]map[string]interface{}
-		expectedNumCalls int
-	}{
-		{
-			name: "happy flow - generate noninterpolated params",
-			baseGenerators: []argoprojiov1alpha1.ApplicationSetNestedGenerator{
-				{ Clusters: interpolatedClusterGenerator },
-				{ Git: nonInterpolatedGitGenerator },
-			},
-			expected: []map[string]interface{}{
-				{
-					"path":                        "examples/git-generator-files-discovery/base.json",
-					"path.basename":               "base",
-					"path.basenameNormalized":     "base",
-					"path.filename":               "base.json",
-					"path.filenameNormalized":     "base.json",
-					"path[0]":                     "examples",
-					"path[1]":                     "git-generator-files-discovery",
-					"path[2]":                     "base",
-					"name":                        "dev-01",
-					"nameNormalized":              "dev-01",
-					"server":                      "https://dev-01.example.com",
-					"metadata.labels.environment": "dev",
-					"metadata.labels.argocd.argoproj.io/secret-type": "cluster",
-				},
-				{
-					"path":                        "examples/git-generator-files-discovery/base.json",
-					"path.basename":               "base",
-					"path.basenameNormalized":     "base",
-					"path.filename":               "base.json",
-					"path.filenameNormalized":     "base.json",
-					"path[0]":                     "examples",
-					"path[1]":                     "git-generator-files-discovery",
-					"path[2]":                     "base",
-					"name":                        "prod-01",
-					"nameNormalized":              "prod-01",
-					"server":                      "https://prod-01.example.com",
-					"metadata.labels.environment": "prod",
-					"metadata.labels.argocd.argoproj.io/secret-type": "cluster",
-				},
-			},
-			clientError: false,
-			expectedMock: [][]map[string]interface{}{
-				{
-					{
-						"path":                        "examples/git-generator-files-discovery/base.json",
-						"path.basename":               "base",
-						"path.basenameNormalized":     "base",
-						"path.filename":               "base.json",
-						"path.filenameNormalized":     "base.json",
-						"path[0]":                     "examples",
-						"path[1]":                     "git-generator-files-discovery",
-						"path[2]":                     "base",
-					},
-				},
-			},
-			expectedNumCalls: 1,
-		},
-		{
-			name: "happy flow - generate interpolated params",
-			baseGenerators: []argoprojiov1alpha1.ApplicationSetNestedGenerator{
-				{ Clusters: interpolatedClusterGenerator },
-				{ Git: interpolatedGitGenerator },
-			},
-			expected: []map[string]interface{}{
-				{
-					"path":                        "examples/git-generator-files-discovery/dev-01.json",
-					"path.basename":               "dev-01",
-					"path.basenameNormalized":     "dev-01",
-					"path.filename":               "dev-01.json",
-					"path.filenameNormalized":     "dev-01.json",
-					"path[0]":                     "examples",
-					"path[1]":                     "git-generator-files-discovery",
-					"path[2]":                     "dev-01",
-					"name":                        "dev-01",
-					"nameNormalized":              "dev-01",
-					"server":                      "https://dev-01.example.com",
-					"metadata.labels.environment": "dev",
-					"metadata.labels.argocd.argoproj.io/secret-type": "cluster",
-				},
-				{
-					"path":                        "examples/git-generator-files-discovery/prod-01.json",
-					"path.basename":               "prod-01",
-					"path.basenameNormalized":     "prod-01",
-					"path.filename":               "prod-01.json",
-					"path.filenameNormalized":     "prod-01.json",
-					"path[0]":                     "examples",
-					"path[1]":                     "git-generator-files-discovery",
-					"path[2]":                     "prod-01",
-					"name":                        "prod-01",
-					"nameNormalized":              "prod-01",
-					"server":                      "https://prod-01.example.com",
-					"metadata.labels.environment": "prod",
-					"metadata.labels.argocd.argoproj.io/secret-type": "cluster",
-				},
-			},
-			clientError: false,
-			expectedMock: [][]map[string]interface{}{
-				{}, // Needed, because the 1st call will fail without interpolation
-				{ // Subsequent calls will succeed with interpolation
-					{
-						"path":                    "examples/git-generator-files-discovery/dev-01.json",
-						"path.basename":           "dev-01",
-						"path.basenameNormalized": "dev-01",
-						"path.filename":           "dev-01.json",
-						"path.filenameNormalized": "dev-01.json",
-						"path[0]":                 "examples",
-						"path[1]":                 "git-generator-files-discovery",
-						"path[2]":                 "dev-01",
-					},
-				},
-				{
-					{
-						"path":                    "examples/git-generator-files-discovery/prod-01.json",
-						"path.basename":           "prod-01",
-						"path.basenameNormalized": "prod-01",
-						"path.filename":           "prod-01.json",
-						"path.filenameNormalized": "prod-01.json",
-						"path[0]":                 "examples",
-						"path[1]":                 "git-generator-files-discovery",
-						"path[2]":                 "prod-01",
-					},
-				},
-			},
-			expectedNumCalls: 3,
-		},
-	}
-	clusters := []client.Object{
-		&corev1.Secret{
-			TypeMeta: metav1.TypeMeta{
-				Kind:       "Secret",
-				APIVersion: "v1",
-			},
-			ObjectMeta: metav1.ObjectMeta{
-				Name:      "dev-01",
-				Namespace: "namespace",
-				Labels: map[string]string{
-					"argocd.argoproj.io/secret-type": "cluster",
-					"environment":                    "dev",
-				},
-			},
-			Data: map[string][]byte{
-				"config": []byte("{}"),
-				"name":   []byte("dev-01"),
-				"server": []byte("https://dev-01.example.com"),
-			},
-			Type: corev1.SecretType("Opaque"),
-		},
-		&corev1.Secret{
-			TypeMeta: metav1.TypeMeta{
-				Kind:       "Secret",
-				APIVersion: "v1",
-			},
-			ObjectMeta: metav1.ObjectMeta{
-				Name:      "prod-01",
-				Namespace: "namespace",
-				Labels: map[string]string{
-					"argocd.argoproj.io/secret-type": "cluster",
-					"environment":                    "prod",
-				},
-			},
-			Data: map[string][]byte{
-				"config": []byte("{}"),
-				"name":   []byte("prod-01"),
-				"server": []byte("https://prod-01.example.com"),
-			},
-			Type: corev1.SecretType("Opaque"),
-		},
-	}
-
-	// convert []client.Object to []runtime.Object, for use by kubefake package
-	runtimeClusters := []runtime.Object{}
-	for _, clientCluster := range clusters {
-		runtimeClusters = append(runtimeClusters, clientCluster)
-	}
-
-	for _, testCase := range testCases {
-		testCaseCopy := testCase // Since tests may run in parallel
-
-		t.Run(testCaseCopy.name, func(t *testing.T) {
-			genMock := &generatorMock{}
-			appSet := &argoprojiov1alpha1.ApplicationSet{}
-
-			appClientset := kubefake.NewSimpleClientset(runtimeClusters...)
-			fakeClient := fake.NewClientBuilder().WithObjects(clusters...).Build()
-			cl := &possiblyErroringFakeCtrlRuntimeClient{
-				fakeClient,
-				testCase.clientError,
-			}
-			var clusterGenerator = NewClusterGenerator(cl, context.Background(), appClientset, "namespace")
-			logrus.Debug(clusterGenerator)
-			for _, g := range testCaseCopy.baseGenerators {
-
-				gitGeneratorSpec := argoprojiov1alpha1.ApplicationSetGenerator{
-					Clusters: g.Clusters,
-					Git:      g.Git,
-				}
-
-				genMock.On("GetTemplate", &gitGeneratorSpec).
-					Return(&argoprojiov1alpha1.ApplicationSetTemplate{})
-			}
-
-			for _, currMock := range testCaseCopy.expectedMock {
-				genMock.On("GenerateParams", mock.AnythingOfType("*v1alpha1.ApplicationSetGenerator"), appSet).Return(currMock, nil).Once()
-			}
-
-			var matrixGenerator = NewMatrixGenerator(
-				map[string]Generator{
-					"Git":      genMock,
-					"Clusters": clusterGenerator,
-				},
-			)
-
-			got, err := matrixGenerator.GenerateParams(&argoprojiov1alpha1.ApplicationSetGenerator{
-				Matrix: &argoprojiov1alpha1.MatrixGenerator{
-					Generators: testCaseCopy.baseGenerators,
-					Template:   argoprojiov1alpha1.ApplicationSetTemplate{},
-				},
-			}, appSet)
-
-			genMock.AssertNumberOfCalls(t, "GenerateParams", testCase.expectedNumCalls)
-
-			if testCaseCopy.expectedErr != nil {
-				assert.ErrorIs(t, err, testCaseCopy.expectedErr)
-			} else {
-				assert.NoError(t, err)
-				assert.Equal(t, testCaseCopy.expected, got)
-			}
-		})
-	}
-}
-
 type generatorMock struct {
 	mock.Mock
 }
@@ -1322,8 +1107,8 @@ func TestGitGenerator_GenerateParams_list_x_git_matrix_generator(t *testing.T) {
 		},
 	}
 
-	repoServiceMock := testutils.ArgoCDServiceMock{Mock: &mock.Mock{}}
-	repoServiceMock.Mock.On("GetFiles", mock.Anything, mock.Anything, mock.Anything, mock.Anything).Return(map[string][]byte{
+	repoServiceMock := &mocks.Repos{}
+	repoServiceMock.On("GetFiles", mock.Anything, mock.Anything, mock.Anything, mock.Anything, mock.Anything).Return(map[string][]byte{
 		"some/path.json": []byte("test: content"),
 	}, nil)
 	gitGenerator := NewGitGenerator(repoServiceMock)
