@@ -12,11 +12,13 @@ import (
 type Context struct {
 	t *testing.T
 	// seconds
-	timeout int
-	name    string
-	project string
-	server  string
-	upsert  bool
+	timeout     int
+	name        string
+	project     string
+	server      string
+	upsert      bool
+	namespaces  []string
+	bearerToken string
 }
 
 func Given(t *testing.T) *Context {
@@ -45,6 +47,11 @@ func (c *Context) Server(server string) *Context {
 	return c
 }
 
+func (c *Context) Namespaces(namespaces []string) *Context {
+	c.namespaces = namespaces
+	return c
+}
+
 func (c *Context) And(block func()) *Context {
 	block()
 	return c
@@ -61,6 +68,11 @@ func (c *Context) Project(project string) *Context {
 	return c
 }
 
+func (c *Context) BearerToken(bearerToken string) *Context {
+	c.bearerToken = bearerToken
+	return c
+}
+
 func (c *Context) Upsert(upsert bool) *Context {
 	c.upsert = upsert
 	return c
